﻿using UnityEngine;
using System.Collections.Generic;
using MeshVR;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using OldMoatGames;
using System;
using OldMoatGames;
using System.Linq;
using Battlehub.RTCommon;

/* /////////////////////////////////////////////////////////////////////////////////////////////////

Author: PluginIdea 
Email:819661664@qq.com
Patreon:https://www.patreon.com/PluginIdea

©2022 PluginIdea, Inc. All rights reserved.

///////////////////////////////////////////////////////////////////////////////////////////////// */

namespace PluginIdea
{
    public class DynamicDecal : MVRScript, IGL
    {

        public struct BlendMode
        {
            public const int Zero = 0;
            public const int One = 1;
            public const int DstColor = 2;
            public const int SrcColor = 3;
            public const int OneMinusDstColor = 4;
            public const int SrcAlpha = 5;
            public const int OneMinusSrcColor = 6;
            public const int DstAlpha = 7;
            public const int OneMinusDstAlpha = 8;
            public const int SrcAlphaSaturate = 9;
            public const int OneMinusSrcAlpha = 10;
        }

        private List<string> blendModes = new List<string> { "NormalBlend", "Multiply" };

        private JSONStorableColor mainColor;
        private JSONStorableFloat mainAlpha;
        private JSONStorableFloat glossiness;
        private JSONStorableUrl abedoTextureUrl;
        private JSONStorableString test;
        private JSONStorableUrl abedoGifTextureUrl;
        private JSONStorableUrl normalTextureUrl;
        private JSONStorableFloat metallic;
        private JSONStorableUrl metallicMapUrl;
        private JSONStorableColor emissionColor;
        private JSONStorableFloat cutoutRadio;
        private JSONStorableFloat occlusion;

        private JSONStorableFloat scaleX;
        private JSONStorableFloat fieldOfView;
        private JSONStorableFloat farClipPlane;

        private JSONStorableUrl metallicGlossMap;
        private JSONStorableUrl emissionMapUrl;
        private JSONStorableUrl occlusionMapUrl;

        private JSONStorableStringChooser blendModeChooser;
        private JSONStorableBool isOrthographic;
        private JSONStorableStringChooser IgnoreLayers;
        private TimerManager timerManager;
        private TextureSettings textureSettings;
        private bool isShowGizmos = true;

        private Dictionary<string, Shader> dic_BundleShaders = new Dictionary<string, Shader>();
        private List<Texture2D> decalTextures;
        private Material defaultMat;

        private float abdeoHWRaido = 1;

        private bool isExternalResourcesLoaded;
        private bool IsExternalResourcesLoaded
        {
            get { return this.isExternalResourcesLoaded; }
        }

        private Material Mat;

        private Projector decalProjector;
        private Projector DecalProjector
        {
            get {
                if (decalProjector == null)
                {
                    CreateProjectorObject();
                }

                return decalProjector;
            }
        }

        public override void Init()
        {
            base.Init();

            if (containingAtom.type != "Empty")
            {
                SuperController.LogError("need add on Empty Atom");
                return;
            }

            timerManager = new TimerManager();
            SuperController.singleton.onSceneLoadedHandlers += HideAllPersonMesh;
            SuperController.singleton.onAtomAddedHandlers += OnAtomAdded;
            textureSettings = new TextureSettings() { wrapMode = TextureWrapMode.Clamp };

            ResourceInit();
            UIInit();
        }

        private void Start()
        {
            GLRenderer.Instance.Add(this);
        }

        #region Init
        private void UIInit()
        {
            Utils.OnInitUI(CreateUIElement);
            Utils.SetupInfoText(this, "<size=40><b>Dynamic Decal(Free) 1.0</b></size>\n\n" +
                "Dynamic Decal is used for real-time decals.\n\n" +
                 "<b>Emission Texture:</b>Use R Pass.\n\n" , 800, true);

            //-------  Color -------------
            //MainColor
            Utils.SetupSpacer(this, 20, true);
            mainColor = Utils.SetupColor(this, "MainColor", Color.white, true);
            mainColor.setCallbackFunction = (h, s, v) =>
            {
                if (!IsExternalResourcesLoaded) return;
                Mat.SetColor("_Color", Color.HSVToRGB(h, s, v));
            };

            emissionColor = Utils.SetupColor(this, "EmissionColor", Color.black, true);
            emissionColor.setCallbackFunction = (h, s, v) =>
            {
                if (!IsExternalResourcesLoaded) return;
                Mat.SetColor("_EmissionColor", Color.HSVToRGB(h, s, v));
            };

            isOrthographic = Utils.SetupToggle(this, "Is Orthographic", true, false);
            isOrthographic.setCallbackFunction = (val) => {
                if (!IsExternalResourcesLoaded) return;
                DecalProjector.orthographic = val;
            };

            // scale
            Utils.SetupSpacer(this, 60, false);
            scaleX = Utils.SetupSliderFloatWithRange(this, "Size X", 0.5f, 0.1f, 5f, false);
            scaleX.setCallbackFunction = (val) =>
            {
                if (!IsExternalResourcesLoaded) return;
                DecalProjector.orthographicSize = val;
            };

            fieldOfView = Utils.SetupSliderFloatWithRange(this, "Field Of View", 30f, 0, 180, false);
            fieldOfView.setCallbackFunction = (val) =>
            {
                if (!IsExternalResourcesLoaded) return;
                DecalProjector.fieldOfView = val;
            };

            farClipPlane = Utils.SetupSliderFloatWithRange(this, "Far Clip Plane", 1, 0.1f, 5, false);
            farClipPlane.setCallbackFunction = (val) =>
            {
                if (!IsExternalResourcesLoaded) return;
                DecalProjector.farClipPlane = val;
            };

            blendModeChooser = Utils.SetupStringChooser(this, "BlendMode", blendModes, false);
            blendModeChooser.setCallbackFunction = (val) => {
                if (!IsExternalResourcesLoaded) return;

                cutoutRadio.val = 0.5f;
                switch (val)
                {
                    case "NormalBlend":
                        Mat.SetInt("_SrcBlend", BlendMode.SrcAlpha);
                        Mat.SetInt("_DstBlend", BlendMode.OneMinusSrcAlpha);
                        cutoutRadio.val = 0;
                        break;
                    case "Multiply":
                        Mat.SetInt("_SrcBlend", BlendMode.DstColor);
                        Mat.SetInt("_DstBlend", BlendMode.Zero);
                        break;
                    default:
                        break;
                }
            };

            ParaMeterSet();
        }

        private void ResourceInit()
        {
            decalTextures = new List<Texture2D>() {
                abedoTex, matallicTex, normalTex, occlusionTex, emssionTex
            };

            HideAllPersonMesh();

            string assetsBundlePath = Utils.GetPluginPath(this) + "/resources/dynamicdecal.assetbundle";
            Request request = new Request { path = assetsBundlePath, callback = OnAssetBundleLoaded };
            AssetLoader.QueueLoadAssetBundleFromFile(request);
        }

        private void OnAssetBundleLoaded(Request request)
        {
            Shader[] shaders = request.assetBundle.LoadAllAssets<Shader>();
            for (int i = 0; i < shaders.Length; i++)
            {
                //SuperController.LogMessage("Shader: " + shaders[i].name);
                if (!dic_BundleShaders.ContainsKey(shaders[i].name))
                {
                    dic_BundleShaders.Add(shaders[i].name, shaders[i]);
                }
            }

            CreateProjectorObject();
        }
        #endregion

        private Texture2D abedoTex, matallicTex, normalTex, occlusionTex, emssionTex;
        private void ParaMeterSet()
        {
            // common UI
            Utils.SetupToggle(this, "Show Gizmos", true, false).setCallbackFunction = (val) =>
            {
                isShowGizmos = val;
            };

            Utils.SetupSpacer(this, 40, false);
            Utils.SetupInfoTextNoScroll(this, "<size=40><b>Common Parameters</b></size>", 45, false);
            Utils.SetupInfoTextNoScroll(this, "<size=30><b>Abedo</b></size>", 35, false);
            abedoTextureUrl = Utils.SetupTexture2DChooser(this, "Set Main Decal Texure", $"Custom/Assets/MyAssets", false, textureSettings, (tex) =>
            {
                if (!IsExternalResourcesLoaded) return;

                if (abedoTex != null) Destroy(abedoTex);
                abedoTex = tex;
                Mat.SetTexture("_MainTex", abedoTex);
                SyncDecalSize();
            });

            mainAlpha = Utils.SetupSliderFloatWithRange(this, "Main Alpha", 1, 0f, 1f, false);
            mainAlpha.setCallbackFunction = (val) =>
            {
                if (!IsExternalResourcesLoaded) return;
                Color color = Mat.GetColor("_Color");
                color.a = val;
                Mat.SetColor("_Color", color);
            };

            Utils.SetupTwinButton(this, "Reload", () => {
                Destroy(Mat.GetTexture("_MainTex"));
                abedoTextureUrl.Reload();
            }, "Remove", () => {
                Destroy(Mat.GetTexture("_MainTex"));
                Mat.SetTexture("_MainTex", defaultMat.GetTexture("_MainTex"));
            }, false);

            // emission
            Utils.SetupSpacer(this, 20, false);
            Utils.SetupInfoTextNoScroll(this, "<size=30><b>Emission</b></size>", 35, false);
            emissionMapUrl = Utils.SetupTexture2DChooser(this, "Set Main EmissionTex", $"Custom/Assets/MyAssets", false, textureSettings, (tex) =>
            {
                if (!IsExternalResourcesLoaded) return;

                if (emssionTex != null) Destroy(emssionTex);
                emssionTex = tex;
                Mat.SetTexture("_EmissionTex", emssionTex);
            });

            Utils.SetupTwinButton(this, "Reload", () => {
                Destroy(Mat.GetTexture("_EmissionTex"));
                emissionMapUrl.Reload();
            }, "Remove", () => {
                Destroy(Mat.GetTexture("_EmissionTex"));
                Mat.SetTexture("_EmissionTex", defaultMat.GetTexture("_EmissionTex"));
            }, false);

            // cutOut
            cutoutRadio = Utils.SetupSliderFloat(this, "Alpha CutOff", 0.5f, 0f, 1f, false);
            cutoutRadio.setCallbackFunction = (val) =>
            {
                if (!IsExternalResourcesLoaded) return;
                Mat.SetFloat("_CutOff", val);
            };
        }

        private void CreateProjectorObject()
        {
            GameObject decalObject = new GameObject("DecalObject");
            decalObject.transform.SetParent(containingAtom.reParentObject.Find("object/rescaleObject"));
            decalObject.transform.localPosition = Vector3.zero;
            decalObject.transform.localRotation = Quaternion.identity;
            decalObject.transform.localScale = Vector3.one;
            decalProjector = decalObject.AddComponent<Projector>();
            decalProjector.nearClipPlane = 0.1f;
            decalProjector.farClipPlane = 1f;
            decalProjector.fieldOfView = 30;
            decalProjector.orthographic = true;
            decalProjector.aspectRatio = abdeoHWRaido;
            decalProjector.orthographicSize = 0.5f;
            decalProjector.ignoreLayers = 1 << LayerMask.NameToLayer("Water");
            Mat = new Material(dic_BundleShaders["PluginIdea/ProjectLight"]);
            defaultMat = new Material(Mat);
            decalProjector.material = this.Mat;
        }

        private void Update()
        {
            timerManager.Update();
        }

        private void SyncDecalSize()
        {
            if (abedoTex == null) return;

            decalProjector.aspectRatio = abdeoHWRaido = (float)abedoTex.width / abedoTex.height;
        }

        private void HideAllPersonMesh()
        {
            timerManager.AddTimer(1.5f, () => {
                // reload
                // abedoTex, matallicTex, normalTex, occlusionTex, emssionTex;
                isExternalResourcesLoaded = true;
                abedoTextureUrl.Reload();
                //metallicMapUrl.Reload();
                //normalTextureUrl.Reload();
                //occlusionMapUrl.Reload();
                emissionMapUrl.Reload();

                // value trigger
                mainAlpha.setCallbackFunction(mainAlpha.val);
                scaleX.setCallbackFunction(scaleX.val);
                fieldOfView.setCallbackFunction(fieldOfView.val);
                farClipPlane.setCallbackFunction(farClipPlane.val);
                //metallic.setCallbackFunction(metallic.val);
                //glossiness.setCallbackFunction(glossiness.val);
                //occlusion.setCallbackFunction(occlusion.val);
                cutoutRadio.setCallbackFunction(cutoutRadio.val);
                mainColor.setCallbackFunction(mainColor.val.H, mainColor.val.S, mainColor.val.V);
                emissionColor.setCallbackFunction(emissionColor.val.H, emissionColor.val.S, emissionColor.val.V);
                blendModeChooser.setCallbackFunction(blendModeChooser.val);

                SuperController.singleton.GetAtoms().Where(x => x.type == "Person").ToList().ForEach(ps =>
                {
                    HidePersonMesh(ps);
                });
            });
        }

        private void OnAtomAdded(Atom atom)
        {
            if (atom.type == "Person")
            {
                // a stupid trick
                timerManager.AddTimer(1f, () => {
                    HidePersonMesh(atom);
                });
            }
        }

        private void HidePersonMesh(Atom person)
        {
            person.GetComponentsInChildren<Renderer>().ToList().ForEach(rd =>
            {
                Utils.SetLayer(rd.transform,LayerMask.NameToLayer("Water"));
            });

            var geo = person.transform.Find("rescale2/geometry");
            Utils.SetLayer(geo, LayerMask.NameToLayer("Water"));
        }

        private void OnPostRender()
        {
            if (isShowGizmos)
            {
                GizmosHelper.DrawProjector(DecalProjector);
                GizmosHelper.DrawArrow3D(DecalProjector.transform, DecalProjector.farClipPlane * 0.6f, Color.green);
            }
        }

        private void OnDestroy()
        {
            if (decalProjector != null)
            {
                Destroy(decalProjector.gameObject);
            }
            timerManager = null;
            SuperController.singleton.onSceneLoadedHandlers -= HideAllPersonMesh;
            SuperController.singleton.onAtomAddedHandlers -= OnAtomAdded;
            GLRenderer.Instance.Remove(this);

            // destroy textures
            decalTextures.ForEach(tex => {
                if (tex != null)
                {
                    Destroy(tex);
                }
            });
            decalTextures.Clear();

            Utils.OnDestroyUI();
        }

        public void Draw(int cullingMask)
        {
            OnPostRender();
        }
    }
}
